########################################################################
#
# File Name: 	        DefiningScope.py
#
# Documentation:	http://docs.ftsuite.com/4ODS/DefiningScope.py.html
#
"""
Implements the DefiningScope meta-data interface.
WWW: http://4suite.com/4ODS         e-mail: support@4suite.com

Copyright (c) 1999 FourThought LLC, USA.   All Rights Reserved.
See  http://4suite.com/COPYRIGHT  for license and copyright information
"""
import string

from Ft.Ods.Collections import LiteralListOfRepositoryObjects
from Ft.Ods.MetaData import MetaObject
from Ft.Ods.MetaData import Scope
from Ft.Ods.StorageManager.Adapters import Constants


class DefiningScope:

    class InvalidType:
        def __init__(self,reason):
            self.reason = reason
    class InvalidExpression:
        def __init__(self,reason):
            self.reason = reason
    class CannotRemove:
        def __init__(self,reason):
            self.reason = reason

    def __init__(self):
        pass
    
    #Override the Scope interface
    def bind(self,name,obj):
        for d in self.defines._4ods_getContents():
            if d.name == name:
                raise Scope.Scope.DuplicateNames(name)
        self.add_defines(obj)

    def resolve(self,name):
        if name[0] == ':':
            from Ft.Ods import Database
            return Database.Database().schema().resolve(name[2:])
        path = string.split(name,'::')
        for d in self.defines._4ods_getContents():
            if d.name == path[0]:
                break
        else:
            raise Scope.Scope.NameNotFound("No attribute %s" % path[0])
        if len(path) == 1:
            return d
        return d.resolve(string.join(path[1:],'::'))

    def unbind(self,name):
        path = string.split(name,'::')
        for d in self.defines._4ods_getContents():
            if d.name == path[0]:
                break
        else:
            raise Scope.Scope.NameNotFound("No attribute %s" % path[0])
        if len(path) == 1:
            self.remove_defines(d)
        return d.unbind(string.join(path[1:],'::'))

    def children(self):
        return self.defines


    def form_defines(self,target,inverse=1):
        self._4ods_formRelationship('defines',LiteralRepositoryObjectList.LiteralRepositoryObjectList,'definedIn','form',target,inverse)

    def add_defines(self, target, inverse=1):
        self._4ods_addRelationship('defines',MetaObject.MetaObject,'definedIn','form',target,inverse)

    def drop_defines(self,target,inverse=1):
        self._4ods_dropRelationship('defines','definedIn','drop',target,inverse)
        

    def remove_defines(self, target, inverse=1):
        self._4ods_removeRelationship('defines','definedIn','drop',target,inverse)


    def create_primitive_type(self,kind):
        from Ft.Ods.MetaData import PrimitiveType
        return PrimitiveType.PrimitiveType(kind)
        

    def create_collection(self,kind,max_size,sub_type):
        from Ft.Ods.MetaData import Collection
        c = Collection.Collection(collection_kind = kind)
        if max_size:
            c.form_max_size(max_size)
        if sub_type:
            c.form_subtype(sub_type)
        return c

    def create_dictionary_type(self,key_type,sub_type):
        from Ft.Ods.MetaData import Dictionary
        dict = Dictionary.Dictionary()
        dict.subtype = sub_type
        dict.key_type = key_type
        return dict

    def create_operand(self, expression):
        from Ft.Ods.Parsers.Odl import OdlParserImp
        p = OdlParserImp.OdlParserImp()
        return p.parseExpression(expression)


    def create_member(self, memberName, memberType):
        from Ft.Ods.MetaData import Member
        m = Member.Member()
        m.name = memberName
        m.form_type(memberType)
        return m

    def create_union_case(self,case_name,case_type,case_labels):
        from Ft.Ods.MetaData import UnionCase
        return UnionCase.UnionCase(case_name,case_type,case_labels)


    def add_constant(self,name,typ,value):
        from Ft.Ods.MetaData import Constant
        c = Constant.Constant()
        c.form_type(typ)
        c.name = name
        c.form_the_Value(value)
        self.add_defines(c)
        return c

    def add_type_definition(self,name,alias):
        from Ft.Ods.MetaData import TypeDefinition
        t = TypeDefinition.TypeDefinition(name,alias)
        self.add_defines(t)
        return t

    def add_enumeration(self,name,element_names):
        from Ft.Ods.MetaData import Enumeration
        from Ft.Ods.MetaData import Constant
        from Ft.Ods.MetaData import Literal

        #NOTE WE don't use add_constant because the values of an
        #enum should not be defined at self's level

        e = Enumeration.Enumeration()
        e.name = name
        ctr = 0
        t = self.resolve('::pk_short')
        eNames = element_names._4ods_getContents()
        for n in eNames:
            l = Literal.Literal()
            l.literal_value = ctr
            c = Constant.Constant()
            c.name = n
            c.form_type(t)
            c.form_the_Value(l)
            e.bind(n,c)
            ctr = ctr + 1
        self.add_defines(e)
        return e

    def add_structure(self,name,fields):
        from Ft.Ods.MetaData import Structure
        s = Structure.Structure()
        s.name = name
        for f in fields:
            s.bind(f.name,f)
        self.add_defines(s)
        return s

    def add_union(self,name,switch_type,cases):
        from Ft.Ods.MetaData import Union
        u = Union.Union(name,switch_type,cases)
        self.add_defines(u)
        return u

    def add_exception(self,name,result):
        from Ft.Ods.MetaData import Exception
        e = Exception.Exception(name,result)
        self.add_defines(e)
        return e

    def remove_constant(self,c):
        #FIXME Where else is a constant references??
        if not c._4ods_checkRemove():
            raise DefiningScope.CannotRemove()
        self.remove_defines(c)
        
    def remove_type_definition(self,t):
        if not t._4ods_checkRemove():
            raise DefiningScope.CannotRemove()
        self.remove_defines(t)

    def remove_enumeration(self,e):
        if not e._4ods_checkRemove():
            raise DefiningScope.CannotRemove()
        self.remove_defines(e)

    def remove_structure(self,s):
        if not s._4ods_checkRemove():
            raise DefiningScope.CannotRemove()
        self.remove_defines(s)

    def remove_union(self,u):
        if not u._4ods_checkRemove():
            raise DefiningScope.CannotRemove()
        self.remove_defines(u)

    def remove_exception(self,e):
        if not e._4ods_checkRemove():
            raise DefiningScope.CannotRemove()
        self.remove_defines(e)




    readOnlyAttributes = ('defines',)
    readWriteAttributes = ()
    relationships = ('defines',)

    tupleNames = ('defines',)
    tupleTypes = (Constants.Types.OBJECT_LIST,)

    _collections = ('defines',)
    collectionMap = {'defines':LiteralListOfRepositoryObjects.LiteralListOfRepositoryObjects
                     }
    
    localExtents = ()
