########################################################################
#
# File Name: 	        PersistentObject.py
#
# Documentation:	http://docs.ftsuite.com/4ODS/RepositoryObject.py.html
#
"""
Implements the Repository meta-data interface.
WWW: http://4suite.com/4ODS         e-mail: support@4suite.com

Copyright (c) 2000 FourThought Inc, USA.   All Rights Reserved.
See  http://4suite.com/COPYRIGHT  for license and copyright information
"""

from Ft.Ods import Database
from Ft.Ods.StorageManager.Adapters import Constants
from Ft.Ods.Collections import CollectionBase
from Ft.Ods import IntegrityException
import tempfile,os

TEMP_DIR = os.environ.get("FTODS_LARGE_FILE_DIR",'/var/local/data/ftods/')
tempfile.tempdir = TEMP_DIR + '/'


class PersistentObject:
    _objectIds = {}
    _literalIds = {}
    def __init__(self):
        self.__dict__['_objectIds'] = {}
        self.__dict__['_collectionIds'] = {}
        self.__dict__['_initialized'] = 0
        self.__dict__['_fileNames'] = {}
        
        #See if we are in a transaction
        self.__dict__['_db'] = Database.Database()
        if self._db._4ods_inTransaction():
            self.__dict__['_tx'] = self._db.current()
            if self._tx:
                self._tx._4ods_registerPersistentObject(self)
        else:
            self.__dict__['_tx'] = None

        #Initialize storage for modified attributes
        self.__dict__['_modifiedAttrs'] = {}


    def _pseudo_del(self):
        #Remove all circular references from relationships
        for name in self.relationships:
            if name not in self._collections:
                if self.__dict__.has_key(name) and self.__dict__[name] != None:
                    #Break this relationship
                    obj = self.__dict__[name]
                    del self.__dict__[name]
                    obj._pseudo_del()
            else:
                if self.__dict__.has_key(name):
                    obj = self.__dict__[name]
                    del self.__dict__[name]
                    obj._pseudo_del()
                    


    def _4ods_getId(self):
        raise "Override please"

    def _4ods_getPersistentType(self):
        raise "Override please"

    def _4ods_setTransaction(self,newTx,unregister=0):
        if self._tx and unregister:
            self._tx._4ods_unregisterPersistentObject(self)
        self.__dict__['_tx'] = newTx
        if self._tx:
            self._tx._4ods_registerPersistentObject(self)

    def delete(self,recurse = 0):
        if self._tx:
            self._tx._4ods_deletePersistentObject(self)
        Types = Constants.Types
        lists = Constants.g_listTypes
        for name in self.relationships:
            i = list(self.tupleNames).index(name)
            typ = self.tupleTypes[i]
            obj = None
            if typ in [Types.OBJECT]:
                obj = getattr(self,name)
                if obj:
                    funcName = 'drop_%s' % name
                    func = getattr(self,funcName)
                    func(self.__dict__[name])
                    if recurse:
                        obj.delete(recurse)
            elif lists[typ]:
                coll = getattr(self,name)
                while 1:
                    if not len(coll._4ods_getContents()):
                        break
                    element = coll._4ods_getContents()[0]
                    funcName = 'remove_%s' % name
                    func = getattr(self,funcName)
                    func(element)
                    if recurse:
                        element.delete(recurse)

        for name in self.readOnlyAttributes + self.readWriteAttributes:
            i = list(self.tupleNames).index(name)
            typ = self.tupleTypes[i]
            if typ == Types.OBJECT:
                obj = getattr(self,name)
                self.__dict__[name] = None
                if self.__dict__['_objectIds'].has_key(name):
                    del self.__dict__['_objectIds'][name]
                if obj:
                    obj.delete(recurse)
            if typ == Types.STRUCTURE:
                struct = getattr(self,name)
                struct.delete()
            if typ == Types.FILE:
                if self.__dict__['_fileNames'].has_key(name):
                    if self.__dict__['_tx']:
                        self.__dict__['_tx']._4ods_registerLargeFileDelete(self.__dict__['_fileNames'][name])
                    del self.__dict__['_fileNames'][name]
            if lists[typ]:
                if not self.__dict__.has_key(name) and not self.__dict__['_collectionIds'].has_key(name):
                    continue
                obj = getattr(self,name)
                obj.delete(recurse)
            

    def __getattr__(self,name):
        if name in self._collections:
            #See if we need to do anything special for the collection
            if self.__dict__['_collectionIds'].has_key(name):
                #Load an onld one
                obj = self._db._4ods_getPersistentObjectById(Constants.ObjectTypes.COLLECTION,self.__dict__['_collectionIds'][name])
                self.__dict__[name] = obj
                return obj
            else:
                #Create a new one
                obj = self.collectionMap[name]()
                self.__dict__[name] = obj
                self.__dict__['_modifiedAttrs'][name] = 1
                return obj
        if self.__dict__['_objectIds'].has_key(name):
            #We have an ID, let's create
            obj = self._db._4ods_getPersistentObjectById(self._4ods_getPersistentType(),self.__dict__['_objectIds'][name])
            self.__dict__[name] = obj
            return obj
        elif self.__dict__['_fileNames'].has_key(name):
            fName = self.__dict__['_fileNames'][name]
            if not fName:
                return ''
            f = open(fName,'r')
            d = f.read()
            f.close()
            self.__dict__[name] = d
            return d
        elif name in self.relationships:
            #A dead object
            return None
        elif name in self.readOnlyAttributes or name in self.readWriteAttributes:
            #This is an unloaded file attribute, or a object attribute
            index = list(self.tupleNames).index(name)
            if self.tupleTypes[index] == Constants.Types.OBJECT:
                return None
            return ''
        if name == 'typeId' and not hasattr(self.__class__,'typeId'):
            return getattr(self.__class__,'_4ods_getTypeId')()
        raise AttributeError(name)

    def __setattr__(self,name,value):
        #check relationships forst because a lot of classes will still have rels in
        #Thier readonly attr list

        if name in self.relationships:
            #This means form!
            if self.__dict__.has_key(name) and isinstance(self.__dict__[name],CollectionBase.CollectionBase):
                raise "Cannot use setattr short cut on N-? relationships"
            func = getattr(self,'form_%s' % name)
            func(value)
        elif name in self.readOnlyAttributes:
            raise "ReadOnly Attribute %s" % name
        elif name in self.readWriteAttributes:
            i = list(self.tupleNames).index(name)
            typ = self.tupleTypes[i]
            self.__dict__[name] = value
            if typ == Constants.Types.FILE:
                if not (self.__dict__['_fileNames'].has_key(name) and self.__dict__['_fileNames'][name]):
                    fileName = tempfile.mktemp()
                    self.__dict__['_fileNames'][name] = fileName
                else:
                    fileName = self.__dict__['_fileNames'][name]
                if self.__dict__['_tx']:
                    self.__dict__['_tx']._4ods_registerLargeFileWrite(fileName,value)
            elif self.__dict__['_objectIds'].has_key(name):
                if value is None:
                    self.__dict__[name] = None
                    del self.__dict__['_objectIds'][name]
                else:
                    self.__dict__['_objectIds'][name] = value._4ods_getId()
            self.__dict__['_modifiedAttrs'][name] = 1
        else:
            self.__dict__[name] = value

    
    def __delattr__(self,name):
        if name in self.relationships:
            #This means drop ro remove
            #Make sure it is loaded
            obj = getattr(self,name)
            if isinstance(obj,CollectionBase.CollectionBase):
                func = getattr(self,'remove_%s' % name)
                for o in self.__dict__[name]._4ods_getContents():
                    func(o)
            else:
                o = getattr(self,name)
                if o:
                    func = getattr(self,'drop_%s' % name)
                    func(o)
            
        elif name in self.tupleNames:
            raise "Cannot delete attributes from a Persistent Object"
        else:
            del self.__dict__[name]

    def _4ods_formRelationship(self,name,targetType,inverseName,inverseType,val,inverse):
        if not isinstance(val,targetType):
            raise IntegrityException(name)
        #Only allowed to be in a relationship with one person
        if getattr(self,name):
            raise IntegrityException(name)
            
        self.__dict__[name] = val
        if inverse:
            if inverseType == 'add':
                val._4ods_addRelationship(inverseName,self.__class__,name,'form',self,0)
            elif inverseType == 'form':
                val._4ods_formRelationship(inverseName,self.__class__,name,'form',self,0)
            else:
                raise "Typo"
        self.__dict__['_modifiedAttrs'][name] = 1


    def _4ods_dropRelationship(self,name,inverseName,inverseType,val,inverse):
        self.__dict__[name] = None
        if self.__dict__['_objectIds'].has_key(name):
            del self.__dict__['_objectIds'][name]
            
        if inverse:
            if inverseType == 'remove':
                val._4ods_removeRelationship(inverseName,name,'drop',self,0)
            elif inverseType == 'drop':
                val._4ods_dropRelationship(inverseName,name,'drop',self,0)
            else:
                raise "Typo"
        self.__dict__['_modifiedAttrs'][name] = 1


    def _4ods_addRelationship(self,name,targetType,inverseName,inverseType,val,inverse):
        if not isinstance(val,targetType):
            raise IntegrityException(name)
        ourColl = getattr(self,name)
        if not ourColl.allows_duplicates() and ourColl.contains_element(val):
            raise IntegrityException(name)
        ourColl.insert_element(val)
        if inverse:
            if inverseType == 'add':
                val._4ods_addRelationship(inverseName,self.__class__,name,'add',self,0)
            elif inverseType == 'form':
                val._4ods_formRelationship(inverseName,self.__class__,name,'add',self,0)
            else:
                raise "Typo %s" % inverseType



    def _4ods_removeRelationship(self,name,inverseName,inverseType,val,inverse):
        ourColl = getattr(self,name)
        if not ourColl.contains_element(val):
            raise IntegrityException(name)
        ourColl.remove_element(val)
        if inverse:
            if inverseType == 'remove':
                val._4ods_removeRelationship(inverseName,name,'remove',self,0)
            elif inverseType == 'drop':
                val._4ods_dropRelationship(inverseName,name,'remove',self,0)
            else:
                raise "Typo"

    def _4ods_initialize(self,data):
        self.__dict__['_initialized'] = 1

        Types = Constants.Types
        lists = Constants.g_listTypes
        primitives = Constants.g_primitiveTypes

        for t,n,v in map(lambda x,y,z: (x,y,z), self.tupleTypes,self.tupleNames,data):

            if t == Types.OBJECT:
                if self.__dict__.has_key(n):
                    del self.__dict__[n]
                if v:
                    self.__dict__['_objectIds'][n] = v
            elif t == Types.ENUMERATION:
                if v == -1:
                    self.__dict__[n] = None
                else:
                    self.__dict__[n] = self._4ods_getPythonClass(n).elements[v]
            elif lists[t]:
                if self.__dict__.has_key(n):
                    del self.__dict__[n]
                if v:
                    self.__dict__['_collectionIds'][n] = v
            elif t == Types.FILE:
                self.__dict__['_fileNames'][n] = v
            elif primitives[t]:
                self.__dict__[n] = v
            else:
                raise "Unknwon Type %s" % t


    def _4ods_getTuple(self):
        if self.__dict__['_initialized']:
            return self._4ods_getModifiedTuple()
        return self._4ods_getFullTuple()

    def _4ods_getFullTuple(self):
        return self._4ods_getModifiedTuple(0)


    def _4ods_getModifiedTuple(self,modCheck=1):

        rt = []

        names = self.tupleNames
        types = self.tupleTypes
        Types = Constants.Types
        lists = Constants.g_listTypes


        #If nothing was modified then we is done!
        if modCheck and not len(self.__dict__['_modifiedAttrs'].values()):
            return None

        for n,t in map(lambda x,y: (x,y),names,types):
            if modCheck and not self.__dict__['_modifiedAttrs'].has_key(n):
                rt.append(None)
            elif modCheck and not self.__dict__['_modifiedAttrs'][n]:
                rt.append(None)
            elif t in [Types.OBJECT]:
                #Use the ID
                if self.__dict__.has_key(n) and self.__dict__[n] != None:
                    rt.append(self.__dict__[n]._4ods_getId())
                elif self.__dict__['_objectIds'].has_key(n):
                    rt.append(self.__dict__['_objectIds'][n])
                else:
                    rt.append(0)
            elif t == Types.ENUMERATION:
                if self.__dict__[n]:
                    rt.append(self.__dict__[n]._v)
                else:
                    rt .append(-1)
            elif lists[t]:
                if self.__dict__.has_key(n) and self.__dict__[n].cardinality() > 0:
                    rt.append(self.__dict__[n]._4ods_getCid())
                elif self.__dict__['_collectionIds'].has_key(n):
                    rt.append(self.__dict__['_collectionIds'][n])
                else:
                    rt.append(None)
            elif t in [Types.STRING,Types.UNSIGNED_SHORT,Types.BOOLEAN,Types.UNSIGNED_LONG,Types.SIGNED_SHORT,Types.DOUBLE,Types.SIGNED_LONG_LONG, Types.SIGNED_LONG, Types.FLOAT]:
                rt.append(self.__dict__[n])
            elif t == Types.FILE:

                if modCheck and not self.__dict__.has_key(n):
                    rt.append('')
                else:
                    if self.__dict__['_fileNames'].has_key(n) and self.__dict__['_fileNames'][n]:
                        rt.append(self.__dict__['_fileNames'][n])
                    else:
                        rt.append('')
                
            else:
                raise "Unknown Type %s on %s %s" % (t,str(self),n)

        if len(rt) == 0:
            raise "Huh?"
            return None
        return tuple(rt)

        

