########################################################################
#
# File Name:      Dbm.py
#
# Documentation:  http://docs.4suite.org/4ODS/StorageManager/Adapters/Dbm.py.html
#
"""
The DBM database back-end adapter.
WWW: http://4suite.com/4ODS         e-mail: support@4suite.com

Copyright (c) 1999 FourThought, Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import string, types

from Ft.Ods import FtodsBindingException
from Ft.Ods import FtodsObjectNotFound
import Constants

import DbmMappings, DbmHelper
from DbmManagement import *
from Ft.Lib import DbmDatabase
from Ft.Lib import Set

########################################
# NOTE:
# Interfaces in this class are those that every driver must support
#############################################

OBJECT_BINDING = 1
EXTENT_BINDING = 2

LEGAL_NAME_CHARS = "01234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_- :/."

class DbAdapter(DbmHelper.DbmHelper):
    def __init__(self, connectString):
        #Create our connection to the database
        self._connect = connectString
        self._db = Connect(connectString)

        # Transaction starts here, if we had them :)
        #self._db.begin()

        self.clear()
        self._init()

    def clear(self):
        self._repoObjects = {}
        self._newRepoObjects = []

        self._objects = {}
        self._collections = {}

        self._pythonClasses = {}
        self._objectClasses = {}

        self._bindings = {}
    
    ################
    # Transactional Interface
    ################

    def commit(self):
        # Save changes
        self._db.commit()
        self.clear()
        self._db = Connect(self._connect)

    def checkpoint(self):
        #Save changes, but don't close
        self._db.commit()
        self.clear()


    def abort(self):
        # Ignore changes
        self._db.rollback()
        self.clear()
        
    def lock(self,repoId,mode):
        # Single user only, for now...
        pass

    def try_lock(self,repoId,mode):
        # Single user only, for now...
        pass
        
    #####################
    # Repository Management
    #####################

    def newRepositoryObjectId(self,ro):
        if ro.meta_kind._v in [0,1,8,15]:
            rid = hash(ro.absolute_name())
            if rid < 0: rid = rid*-1
        else:
            system = self._db.getTable('ftods_system')
            rid = system['ftods_repoid']
            system['ftods_repoid'] = rid + 1

        ro._4ods_setRepositoryId(rid)
        tableName = DbmMappings.g_metaKindToTableName[ro.meta_kind._v]
        
        self._db[tableName][str(rid)] = None
        self._db['ftods_repomap'][str(rid)] = (rid, ro.meta_kind._v)
        self._repoObjects[rid] = ro
        self._4ods_updateExtentsAdd(ro)
        return rid

    def newRepositoryObject(self, ro):
        if ro.meta_kind._v in [8,1]:
            #Create a new storage class
            self._4ods_createInterfaceStorage(ro)
        if ro.meta_kind._v == 2 and ro.collection_kind._v == 4:
            self._4ods_createDictionaryStorage(ro)
        self._newRepoObjects.append(ro._4ods_getRepositoryId())

    def writeRepositoryObject(self, ro):
        if ro.meta_kind._v in [8,1]:
            #Update the storage if needed
            self._4ods_updateInterfaceStorage(ro)

        rid = ro._4ods_getRepositoryId()
        tableName = DbmMappings.g_metaKindToTableName[ro.meta_kind._v]
        self._db[tableName][str(rid)] = ro._4ods_getFullTuple()
        return 1

    def getRepositoryObject(self, rid):
        if self._repoObjects.has_key(rid):
            return self._repoObjects[rid]
        #Load it
        rt = self._db['ftods_repomap'].get(str(rid))
        if rt is None:
            return
        meta_kind = rt[1]
        instance = DbmMappings.g_metaKindToRepoClass[meta_kind]()

        #Now load the tuple
        tableName = DbmMappings.g_metaKindToTableName[meta_kind]
        tup = self._db[tableName][str(rid)]
        instance._4ods_initialize(tup)
        self._repoObjects[id] = instance
        return instance

    def deleteRepositoryObject(self, ro):
        if ro.meta_kind._v in [8,1]:
            #Update the storage if needed
            self._4ods_deleteInterfaceStorage(ro)
        rid = ro._4ods_getRepositoryId()
        tableName = DbmMappings.g_metaKindToTableName[ro.meta_kind._v]
        del self._db[tableName][str(rid)]
        self._4ods_updateExtentsDelete(ro)

    ###########################
    #Binding and extent interface
    ###########################

    def bind(self, object, name):
        #Make sure the name does not have any illegal characters
        if not type(name) == types.StringType:
            raise FtodsBindingException(name)
        if len(filter(lambda x, lc=LEGAL_NAME_CHARS:x not in lc,name)):
            raise FtodsBindingException(name)

        #Make sure the name is unique
        if self.lookup(name):
            raise FtodsBindingException(name)

        #Add it to the databases unless it is a new object
        oid = object._4ods_getId()
        if oid:
            try:
                # What about the repoid??
                self._db['ftods_binding'][name] = oid
            except:
                raise FtodsBindingException(name)

        #Add this to the binding dict
        self._bindings[name] = object

    def unbind(self,name):
        #Make sure this name exists in the bindings
        if self._bindings.has_key(name):
            del self._bindings[name]
        if self._db['ftods_binding'].has_key(name):
            del self._db['ftods_binding'][name]

    def lookup(self,name):

        if self._bindings.has_key(name):
            return self._bindings[name]

        #Try to load the binding
        rt = self._db['ftods_binding'].get(name)
        if rt:
            oid = rt
            rt = self.getObject(oid)
            self._bindings[name] = rt
        return rt

    def loadExtentNames(self):
        return self._db['ftods_extents'].values()

    def getExtentTuple(self,name):
        ext = self.lookup(name)
        
    def getAllBindings(self,object):
        if object is None:
            return self._db['ftods_binding'].keys()
        else:
            res = []
            for k,o in map(lambda x:(x[0],x[1]._4ods_getId()),self._bindings.items()) + self._db['ftods_binding'].items():
                if o == object._4ods_getId():
                    res.append(k)
            return Set.Unique(res)


    def isBound(self,obj):
        bindings = self._db['ftods_binding'].items()
        oid = obj._4ods_getId()
        for (name, objid) in bindings:
            if objid == oid:
                return name
        return ''

    ############################
    #Collection Interface
    ###########################
    def newCollectionId(self, collection):
        cid = self._4ods_nextId('collection')
        collection._4ods_setCid(cid)
        subType = collection._4ods_getSubType()
        cType = collection.collectionType
        #Store its class
        classId = self._4ods_getPythonClassId(collection.__class__)
        self._db['ftods_collectionclass'][str(cid)] = (cid, classId, subType, cType)

    def writeCollection(self, collection):
        cid = collection._4ods_getCid()
        cType = collection.collectionType
        table_name = DbmMappings.g_collectionTableMapping[cType]
        table = self._db.getTable(table_name)
        values = table.get(str(cid), [])
        for change in collection._4ods_getChanges():
            if change[0] in [Constants.CollectionChanges.INSERT,
                             Constants.CollectionChanges.APPEND]:
                value = change[1]
                index = change[2]
                if change[0] == Constants.CollectionChanges.INSERT:
                    values.insert(index, value)
                else:
                    values.append(value)
            elif change[0] in [Constants.CollectionChanges.REMOVE,
                               Constants.CollectionChanges.REMOVE_END]:
                index = change[1]
                del values[index]
            else:
                raise Exception("Unknown collection change")
        table[str(cid)] = values
        return 1

    def getCollection(self, cid):
        if self._collections.has_key(cid):
            return self._collections[cid]

        #Load it
        pClass,subType,cType = self._4ods_getCollectionClass(cid)
        instance = pClass()
        instance._4ods_setCid(cid)

        table_name = DbmMappings.g_collectionTableMapping[cType]
        size = len(self._db[table_name][str(cid)])
        instance._4ods_initialize(size, subType)
        self._collections[cid] = instance
        return instance

    def loadCollectionData(self, c, index):
        cid = c._4ods_getCid()
        cType = c.collectionType
        table_name = DbmMappings.g_collectionTableMapping[cType]
        items = self._db[table_name].get(str(cid), [])
        if index:
            # Is the index 1 or 0 based??
            return [items[index]]
        else:
            return items
            
    def deleteCollection(self,c):
        cid = str(c._4ods_getCid())
        cType = c.collectionType
        table_name = DbmMappings.g_collectionTableMapping[cType]
        del self._db['ftods_collectionclass'][cid]
        del self._db[table_name][cid]

    #############################
    #Literal Interface
    #############################

    # FIXME: ?????

    #############################
    #Object Interface
    #############################

    def newObjectId(self, obj):
        boundNames = []
        for (name, bound) in self._bindings.items():
            if obj == bound:
                boundNames.append(name)

        oid = self._4ods_nextId('object')
        obj._4ods_setOid(oid)

        classId = self._4ods_getPythonClassId(obj.__class__)

        self._objects[oid] = obj

        self._db['ftods_actualclass'][str(oid)] = (oid, classId)

        self._4ods_updateExtentsAdd(obj)

        binding = self._db['ftods_binding']
        for name in boundNames:
            try:
                # FIXME: what object repoid???
                binding[name] = oid
            except:
                raise FtodsBindingException(name)
        return oid

    def writeObject(self,o):
        oid = o._4ods_getOid()
        tableName = 'interface__%i' % o.typeId
        self._db[tableName][str(oid)] = o._4ods_getFullTuple()
        return 1
        
    def getObject(self,id):
        id = int(id)
        if self._objects.has_key(id):
            return self._objects[id]

        #Load it
        objectClass = self._4ods_getObjectClass(id)
        if not objectClass:
            raise FtodsObjectNotFound(id)

        instance = objectClass()
        repoId = objectClass.typeId

        #Now load the tuple
        tup = self._4ods_readObjectTuple(id, repoId, instance.tupleTypes)
        instance._4ods_initialize(tup)
        instance._4ods_setOid(id)
        self._objects[id] = instance
        return instance

    def deleteObject(self,obj):
        oid = obj._4ods_getId()
        table_name = "interface__%i" % obj.typeId
        del self._db[table_name][str(oid)]
        self._4ods_updateExtentsDelete(obj)

        #Remove it from the bindings
        name = self.isBound(obj)
        name and self.unbind(name)
        
        del self._db['ftods_actualclass'][str(oid)]

    def getAllObjectIds(self):
        oids = self._db['ftods_actualclass'].keys()
        return map(lambda oid: int(oid), oids)

    #Garbage Collection
    def collectGarbage(self):
        self._4ods_collectRepositoryObjectGarbage()
