########################################################################
#
# File Name: 	        Postgres.py
#
# Documentation:	http://docs.ftsuite.com/4ODS/Drivers/Postgres.py.html
#
"""
The Postgres back-end.
WWW: http://4suite.com/4ODS         e-mail: support@4suite.com

Copyright (c) 1999 FourThought, Inc, USA.   All Rights Reserved.
See  http://4suite.com/COPYRIGHT  for license and copyright information
"""
import sys,string,types

import string, threading
from Ft.Ods import FtodsBindingException
from Ft.Ods import FtodsObjectNotFound

from Ft.Ods.StorageManager.Adapters import Constants

import PostgresHelper
import PostgresMappings
import PostgresManagement

########################################
#NOTE:
#Interfaces in this class are those that every driver must support
#############################################


OBJECT_BINDING = 1
EXTENT_BINDING = 2

LEGAL_NAME_CHARS = "01234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_- :/."

class DbAdapter(PostgresHelper.PostgresHelper):    
    def __init__(self, connString):
        #Create our connection to the database

        self._connString = connString
      	self._db = Connect(connString)
        self._db.execute('BEGIN')

        self.clear()
        self._init()


    def clear(self):
        self._repoObjects = {}
        self._newRepoObjects = []

        self._objects = {}
        self._collections = {}

        self._pythonClasses = {}
        self._objectClasses = {}

        self._bindings = {}


    
    ################
    #The Transactional Interface
    ################
    def commit(self):
        self._db.execute('COMMIT')
        self._db = None
        self.clear()

    def abort(self):
        self._db = None        
        self.clear()

    def checkpoint(self):
        self._db.execute('COMMIT')
        self._db.execute('BEGIN')
        

    def lock(self,repoId,mode):
        table_name = self._4ods_getTableNameFromRepoId(txId,repoId)
        if mode == Constants.READ_LOCK:
            #Do a read lock
            qstr = 'LOCK %s IN ACCESS SHARE MODE'
        else:
            #Do a write lock
            qstr = 'LOCK %s in ACCESS EXCLUSIVE MODE'

        self._db.execute(qstr)
        return

    def try_lock(self,repoId,mode):
        #FIXME how do we do this in Postgres???
        self.lock(repoId,mode)

    #####################
    #Repository Management
    #####################
    def newRepositoryObjectId(self,ro):
        #rid = self._db.queryAllTuple("SELECT NEXTVAL('ftods_repoid')")[0][0]
        
        if ro.meta_kind._v in [0,1,8,15]:
            rid = hash(ro.absolute_name())
            if rid < 1:
                rid = rid *-1
        else:
            rid = int(self._db.queryAllTuple("SELECT NEXTVAL('ftods_repoid')")[0][0])
            
        ro._4ods_setRepositoryId(rid)
        tableName = PostgresMappings.g_metaKindToTableName[ro.meta_kind._v]
        qstr = "INSERT INTO %s (_repoid) VALUES(%i)"%(tableName,rid)
        self._db.execute(qstr)
        qstr = "INSERT INTO ftods_repomap (repoid,meta_kind) VALUES(%i,%i)"%(rid,ro.meta_kind._v)
        self._db.execute(qstr)
        self._repoObjects[rid] = ro
        self._4ods_updateExtentsAdd(ro)
        return rid

    def newRepositoryObject(self,ro):
        if ro.meta_kind._v in [8,1]:
            #Create a new storage class
            self._4ods_createInterfaceStorage(ro)
        if ro.meta_kind._v == 2 and ro.collection_kind._v == 4:
            self._4ods_createDictionaryStorage(ro)
        self._newRepoObjects.append(ro._4ods_getRepositoryId())

    def writeRepositoryObject(self,ro):
        values = ro._4ods_getTuple()
        if not values:
            return 0

        if ro.meta_kind._v in [8,1]:
            #Update the storage if needed
            self._4ods_updateInterfaceStorage(ro)

        types = ro.tupleTypes
        names = ro.tupleNames

        qstr = self._4ods_tupleToUpdateString(types,names,values)
        if not qstr: return
        tableName = PostgresMappings.g_metaKindToTableName[ro.meta_kind._v]

        qstr = 'UPDATE %s SET %s WHERE _repoid = %i' % (tableName,
                                                        qstr,
                                                        ro._4ods_getRepositoryId()
                                                        )
        self._db.execute(qstr)
        
        return 1

    def getRepositoryObject(self,id):
        if self._repoObjects.has_key(id):
            return self._repoObjects[id]
        #Load it

        qstr = "SELECT meta_kind FROM ftods_repomap WHERE repoid = %i"%id
        rt = self._db.queryAllTuple(qstr)
        if len(rt) == 0:
            return None
        rt = rt[0][0]
        inst = PostgresMappings.g_metaKindToRepoClass[rt]()

        #Now load the tuple
        tableName = PostgresMappings.g_metaKindToTableName[rt]
        qstr = "SELECT * from %s where _repoid = %s"%(tableName,id)
        tup = self._db.queryAllTuple(qstr)[0]
        tup = self._4ods_toOdmgValues(tup,inst.tupleTypes)
        inst._4ods_initialize(tup)
        self._repoObjects[id] = inst
        return inst

    def deleteRepositoryObject(self,ro):
        if ro.meta_kind._v in [8,1]:
            #Update the storage if needed
            self._4ods_deleteInterfaceStorage(ro)
        tableName = PostgresMappings.g_metaKindToTableName[ro.meta_kind._v]
        qstr ="DELETE from %s where _repoid = %s" %(tableName,ro._4ods_getRepositoryId())
        self._db.execute(qstr)
        self._4ods_updateExtentsDelete(ro)



    ###########################
    #Binding and extent interface
    ###########################
    def bind(self,object,name):

        #Make sure the name does not have any illegal characters
        if not type(name) == types.StringType:
            raise FtodsBindingException(name)
        if len(filter(lambda x, lc=LEGAL_NAME_CHARS:x not in lc,name)):
            raise FtodsBindingException(name)

        #Make sure the name is unique
        if self.lookup(name):
            raise FtodsBindingException(name)

        #Add it to the databases unless it is a new object
        if object._4ods_getId() != 0:
            try:
                self._db.execute("INSERT INTO ftods_binding (name,objid) VALUES ('%s',%d)" % (name,object._4ods_getId()))
            except:
                raise FtodsBindingException(name)

        #Add this to the binding dict
        self._bindings[name] = object
        

    def unbind(self,name):

        #Make sure this name exists in the bindings
        if self._bindings.has_key(name):
            del self._bindings[name]
        self._db.execute("DELETE FROM ftods_binding where name = '%s'" % name)

    def lookup(self,name):
        if self._bindings.has_key(name):
            return self._bindings[name]

        #Try to load the binding
        rt = self._db.queryAllTuple("SELECT objid from ftods_binding where name = '%s'" % name)
        if len(rt):
            #Tis a binding
            oid = rt[0][0]
            obj = self.getObject(long(oid))
            self._bindings[name] = obj
            return obj
        return None

    

    def loadExtentNames(self):
        rt = self._db.queryAllTuple("SELECT name,extentid,persistenttype from ftods_extents")
        return rt

    def getExtentTuple(self,name):
        ext = self.lookup(name)
        

        
    def getAllBindings(self,object):
        cmd = "SELECT name from ftods_binding"
        if object is not None:
            cmd = cmd + " where objid = '%d'" % (object._4ods_getId())
        rt = self._db.queryAllTuple(cmd)
        res =  map(lambda x:x[0],rt)
        for k,o in self._bindings.items():
            if o == object and k not in res:
                res.append(k)
        return res

    def isBound(self,obj):
        rt = self._db.queryAllTuple("SELECT name from ftods_binding where objid = %d" % obj._4ods_getId())
        if rt:
            return rt[0][0]
        return ''


    ############################
    #Collection Interface
    ###########################
    def newCollectionId(self,c):
        cid = self._db.queryAllTuple("SELECT NEXTVAL('ftods_collectionid')")[0][0]
        c._4ods_setCid(cid)
        subType = c._4ods_getSubType()
        #Store its class
        pClassId = self._4ods_getPythonClassId(c.__class__)
        self._db.execute("INSERT INTO ftods_collectionclass (_cid,pythonclassid,subtype) VALUES (%i,%i,%i)"%(cid,pClassId,subType))

    def writeCollection(self,c):

        #Get a list of changes from the collection
        changes = c._4ods_getChanges()
        subType = c._4ods_getSubType()
        collType = c.collectionType
        cid = c._4ods_getCid()

        table_name = PostgresMappings.g_collectionTableMapping[subType][collType]

        for change in changes:
            if change[0] in [Constants.CollectionChanges.INSERT,Constants.CollectionChanges.APPEND]:
                value = change[1]
                index = change[2]
                if change[0] == Constants.CollectionChanges.INSERT:
                    #Shift all other indexs
                    cmd = 'UPDATE %s set index = index + 1 WHERE _cid = %i and index >= %s' % (table_name,cid,index)
                    self._db.execute(cmd)

                cmd = 'INSERT INTO %s (_cid,index,value) VALUES (%i,%i,%s)' % (table_name,cid,index,self._4ods_toSqlValues((value,),(subType,))[0])
                self._db.execute(cmd)
            elif change[0] in [Constants.CollectionChanges.REMOVE,Constants.CollectionChanges.REMOVE_END]:
                index = change[1]

                cmd = 'DELETE FROM %s WHERE _cid = %i and index = %i' % (table_name,cid,index)
                self._db.execute(cmd)

                if change[0] == Constants.CollectionChanges.REMOVE:
                    #Shift all other indexs
                    cmd = 'UPDATE %s set index = index - 1 WHERE _cid = %s and index >= %s' % (table_name,cid,index)
                    self._db.execute(cmd)


                #Shift all other indexes
                
            else:
                raise "Unknown Change"

        return 1

    def getCollection(self,id):
        if self._collections.has_key(id):
            return self._collections[id]
        #Load it

        pClass,subType = self._4ods_getCollectionClass(id)
        inst = pClass()
        inst._4ods_setCid(id)

        collType = inst.collectionType
        table_name = PostgresMappings.g_collectionTableMapping[subType][collType]

        #Now load just the size
        siz = self._db.queryAllTuple("SELECT COUNT(index) from %s where _cid = %i" % (table_name,id))
        inst._4ods_initialize(siz[0][0],subType)
        self._collections[id] = inst
        return inst  

    def loadCollectionData(self,c,index):

        subType = c._4ods_getSubType()
        collType = c.collectionType
        cid = c._4ods_getCid()
        table_name = PostgresMappings.g_collectionTableMapping[subType][collType]
        
        st = "SELECT value from %s where _cid = %i" % (table_name,cid)

        if index:
            st = st + ' and index = %i' % index
        else:
            st = st + ' ORDER BY index'

        rt = self._db.queryAllTuple(st)
        return map(lambda x,f=self._4ods_toOdmgValues,t=(subType,):f((x[0],),t)[0],rt)
            
    def deleteCollection(self,c):

        subType = c._4ods_getSubType()
        collType = c.collectionType
        cid = c._4ods_getCid()
        table_name = PostgresMappings.g_collectionTableMapping[subType][collType]
        

        self._db.execute("DELETE from ftods_collectionclass where cid = %i" % cid)
        self._db.execute("DELETE from %s where cid = %i" % (table_name,cid))
        



    #############################
    #Literal Interface
    #############################


    #############################
    #Object Interface
    #############################
    def newObjectId(self,obj):

        boundNames = []
        for name,bobj in self._bindings.items():
            if obj == bobj:
                boundNames.append(name)


        oid = int(self._db.insert("INSERT INTO interface__%i (dummy) VALUES(0)"%(obj.typeId)))

        obj._4ods_setOid(oid)

        pClassId = self._4ods_getPythonClassId(obj.__class__)

        self._objects[oid] = obj

        self._db.execute("INSERT INTO ftods_actualclass (_oid,pythonclassid) VALUES (%i,%i)"%(oid,pClassId))

        self._4ods_updateExtentsAdd(obj)

        for name in boundNames:
            try:
                self._db.execute("INSERT INTO ftods_binding (name,objid) VALUES ('%s',%d)" % (name,oid))
            except:
                raise FtodsBindingException(name)

        return oid

    def writeObject(self,o):

        values = o._4ods_getTuple()
        if not values:
            return 0

        names = o.tupleNames

        #we need to translate the names to our version
        newNames = []
        ids = o.tupleIds
        newNames = map(lambda x,y: (y and 'def__%s'%y) or x,names,ids)

        types = o.tupleTypes

        qstr = self._4ods_tupleToUpdateString(types,newNames,values)
        if not qstr: return
        tableName = 'interface__%i' % o.typeId

        qstr = 'UPDATE %s SET %s WHERE oid = %i' % (tableName,
                                                    qstr,
                                                    o._4ods_getOid()
                                                    )
        self._db.execute(qstr)
        return 1
        
    def getObject(self,id):
        if self._objects.has_key(id):
            return self._objects[id]
        #Load it


        pClass = self._4ods_getObjectClass(id)
        if not pClass:
            raise FtodsObjectNotFound(id)

        inst = pClass()
        repoId = pClass.typeId

        #Now load the tuple
        tup = self._4ods_readObjectTuple(id,repoId,inst.tupleTypes)
        inst._4ods_initialize(tup)
        inst._4ods_setOid(id)
        self._objects[id] = inst
        return inst        

    def deleteObject(self,obj):
        id = obj._4ods_getId()
        table_name = "interface__%d" % obj.typeId
        self._db.execute("DELETE FROM %s where oid = %d" % (table_name,id))
        self._4ods_updateExtentsDelete(obj)

        #Remove it from the bindings
        self._db.execute("DELETE FROM ftods_binding where objid = %d" % obj._4ods_getId())
        for name in self._bindings.keys():
            if self._bindings[name] == obj:
                del self._bindings[name]
        
        self._db.execute("DELETE FROM ftods_actualclass where _oid = %i"%(id))

    def getAllObjectIds(self):
        rt = self._db.queryAllTuple("SELECT _oid from ftods_actualclass")
        return map(lambda x:int(x[0]),rt)

    #Garbage Collection
    def collectGarbage(self):
        self._4ods_collectRepositoryObjectGarbage()


ClearDb = PostgresManagement.ClearDb
CleanDb = PostgresManagement.CleanDb
DestroyDb = PostgresManagement.DestroyDb
CreateDb = PostgresManagement.CreateDb
GrantAccess = PostgresManagement.GrantAccess
Connect = PostgresManagement.Connect
ExistsDb = PostgresManagement.ExistsDb
