#!/usr/bin/env python
############################################################################
#
# File Name: 		Transaction.py
#
# Documentation:	http://docs.ftsuite.com/4ODS/Transaction.py.html
#


import threading,os

from Ft.Ods.StorageManager.Adapters import Constants
import Ft.Ods


#Exception classes
class TransactionInProgress:
    pass

class TransactionNotInProgress:
    pass
    
class Transaction:
    def __init__(self,factory,sm):
        self.__id = id(self)
        self.__factory = factory
        self.__open = 0
        self.__hasBeenClosed = 0
        self.__lk = threading.RLock()
        self.__storageManager = sm
        self._clear()


    def _clear(self):

        self.__extentsAdd = {}
        self.__extentsDel = {}
        self.__modifiedRepositoryObjects = []
        self.__modifiedObjects = []
        self.__modifiedLiterals = []
        self.__deletedRepositoryObjects = []
        self.__deletedObjects = []
        self.__deletedLiterals = []
        self.__modifiedCollections = []
        self.__deletedCollections = []
        self.__largeFilesToDelete = []
        self.__largeFilesToWrite = []


    def begin(self):
        self.__lk.acquire()
        try:
            if self.__open or self.__hasBeenClosed:
                raise TransactionInProgress()
            self.__open = 1
            self._clear()
        finally:
            self.__lk.release()

    def abort(self):
        self.__lk.acquire()
        try:
            if not self.__open:
                raise TransactionNotInProgress()
            self.__factory._4ods_endTransaction(self.__id)
            self.__storageManager._4ods_abort()
            self.__open = 0
            self.__hasBeenClosed = 1
            self.__storageManager = None
            for ro in self.__modifiedRepositoryObjects + self.__deletedRepositoryObjects:
                Ft.Ods.FreePersistentObject(ro)

            self._clear()
        finally:
            self.__lk.release()

    #Called by the database to any outstanding transactions
    #When the database gets closed
    def _4ods_close(self):
        self.__lk.acquire()
        try:
            self.__open = 0
            self.__hasBeenClosed = 1
            self.__storageManager = None
            for ro in self.__modifiedRepositoryObjects + self.__deletedRepositoryObjects:
                Ft.Ods.FreePersistentObject(ro)
        finally:
            self.__lk.release()

    def commit(self):
        self.__lk.acquire()
        try:
            self.__writeObjects()
            self._clear()
            self.__open = 0
            self.__hasBeenClosed = 1
            self.__storageManager._4ods_commit()
            self.__storageManager = None
        finally:
            self.__lk.release()
        self.__lk.acquire()
        try:
            self.__factory._4ods_endTransaction(self.__id)
        finally:
            self.__lk.release()

    def checkpoint(self):
        self.__lk.acquire()
        try:
            self.__writeObjects()
            self.__storageManager._4ods_checkpoint()
            mos = self.__modifiedRepositoryObjects + self.__modifiedObjects + self.__modifiedLiterals + self.__modifiedCollections
            dos = self.__deletedRepositoryObjects + self.__deletedObjects + self.__deletedLiterals + self.__deletedCollections

            self._clear()

            for o in mos:
                o._4ods_setTransaction(self)
            for o in dos:
                o._4ods_setTransaction(None)
                

        finally:
            self.__lk.release()
        self.__lk.acquire()


    def join(self):
        self.__lk.acquire()
        try:
            if self.__open:
                self.__factory._4ods_registerThread(self)
            else:
                raise TransactionNotInProgress()
        finally:
            self.__lk.release()

    def leave(self):
        self.__lk.acquire()
        try:
            if self.__open:
                self.__factory._4ods_unregisterThread(self)
            else:
                raise TransactionNotIsProgress()
        finally:
            self.__lk.release()

    def __writeObjects(self):
        self.__lk.acquire()
        try:
            if not self.__open:
                raise TransactionNotInProgress()

            #############
            #First go through an give all new objects an ID
            #then write "new" repo objects
            #Then write all objects
            #Then delete all to be deleted objects


            #Check for new objects
            newRepos = filter(lambda x: not x._4ods_getRepositoryId(),self.__modifiedRepositoryObjects)
            newCollections = filter(lambda x: not x._4ods_getCid(),self.__modifiedCollections)
            newLiterals = filter(lambda x: not x._4ods_getLid(),self.__modifiedLiterals)
            newObjects = filter(lambda x: not x._4ods_getOid(),self.__modifiedObjects)

            for ro in newRepos:
                self.__storageManager.newRepositoryObjectId(ro)
            for c in newCollections:
                if not c.cardinality():
                    continue
                self.__storageManager.newCollectionId(c)
            for l in newLiterals:
                self.__storageManager.newLiteralId(l)
            for o in newObjects:
                self.__storageManager.newObjectId(o)


            #We need to tell the DB about new repo objects
            for ro in newRepos:
                self.__storageManager.newRepositoryObject(ro)
                
            #Write all changes
            for ro in self.__modifiedRepositoryObjects:
                if ro in self.__deletedRepositoryObjects:
                    continue
                self.__storageManager.writeRepositoryObject(ro)
            for o in self.__modifiedObjects:
                if o in self.__deletedObjects:
                    continue
                self.__storageManager.writeObject(o)
            for l in self.__modifiedLiterals:
                if l in self.__deletedLiterals:
                    continue
                self.__storageManager.writeLiteral(l)
            for c in self.__modifiedCollections:
                if not c._4ods_isModified():
                    continue
                if c in self.__deletedCollections:
                    continue
                self.__storageManager.writeCollection(c)

            for ro in self.__deletedRepositoryObjects:
                self.__storageManager.deleteRepositoryObject(ro)
            for o in self.__deletedObjects:
                self.__storageManager.deleteObject(o)
            for l in self.__deletedLiterals:
                self.__storageManager.deleteLiteral(l)
            for c in self.__deletedCollections:
                self.__storageManager.deleteCollection(c)

            for fileName in self.__largeFilesToDelete:
                os.unlink(fileName)

            for fileName,data in self.__largeFilesToWrite:
                f = open(fileName,'w')
                f.write(data)
                f.close()

                
            self.__storageManager.collectGarbage()

            for ro in self.__modifiedRepositoryObjects + self.__deletedRepositoryObjects:
                Ft.Ods.FreePersistentObject(ro)


        finally:
            self.__lk.release()

    def isOpen(self):
        return self.__open

    ### Internal Callback Methods ###

    def _4ods_getId(self):
        return self.__id


    def _4ods_registerPersistentObject(self,po):
        self.__lk.acquire()
        try:
            if not self.__open:
                raise TransactionNotInProgress()
            typ = po._4ods_getPersistentType()
            if typ == Constants.ObjectTypes.REPO:
                self.__modifiedRepositoryObjects.append(po)
            elif typ == Constants.ObjectTypes.OBJECT:
                self.__modifiedObjects.append(po)
            elif typ == Constants.ObjectTypes.LITERAL:
                self.__modifiedLiterals.append(po)
            elif typ == Constants.ObjectTypes.COLLECTION:
                self.__modifiedCollections.append(po)
            else:
                raise "Unknown Object Type %s" % typ
            for eName in po._extents:
                if not self.__extentsAdd.has_key(eName):
                    self.__extentsAdd[eName] = []
                self.__extentsAdd[eName].append(po)
        finally:
            self.__lk.release()

    def _4ods_unregisterPersistentObject(self,po):
        self.__lk.acquire()
        try:
            if not self.__open:
                raise TransactionNotInProgress()
            typ = po._4ods_getPersistentType()
            if typ == Constants.ObjectTypes.REPO:
                while po in self.__modifiedRepositoryObjects:
                    index = self.__modifiedRepositoryObjects.index(po)
                    del self.__modifiedRepositoryObjects[index]
            elif typ == Constants.ObjectTypes.OBJECT:
                while po in self.__modifiedObjects:
                    index = self.__modifiedObjects.index(po)
                    del self.__modifiedObjects[index]
            elif typ == Constants.ObjectTypes.LITERAL:
                while po in self.__modifiedLiterals:
                    index = self.__modifiedLiterals.index(po)
                    del self.__modifiedLiterals[index]
            elif typ == Constants.ObjectTypes.COLLECTION:
                while po in self.__modifiedCollections:
                    index = self.__modifiedCollections.index(po)
                    del self.__modifiedCollections[index]
            else:
                raise "Unknown Object Type %s" % typ

            for eName in po._extents:
                if not self.__extentsAdd.has_key(eName) and po in self.__extentsAdd[eName]:
                    index = self.__extentsAdd[eName].index(po)
                    del self.__extentsAdd[eName][index]

        finally:
            self.__lk.release()

    def _4ods_deletePersistentObject(self, po):
        self.__lk.acquire()
        try:
            if not self.__open:
                raise TransactionNotInProgress()
            if not po._4ods_getId():
                return
            typ = po._4ods_getPersistentType()
            if typ == Constants.ObjectTypes.REPO:
                if not po in self.__deletedRepositoryObjects:
                    self.__deletedRepositoryObjects.append(po)
            elif typ == Constants.ObjectTypes.OBJECT:
                if not po in self.__deletedObjects:
                    self.__deletedObjects.append(po)
            elif typ == Constants.ObjectTypes.LITERAL:
                if not po in self.__deletedLiterals:
                    self.__deletedLiterals.append(po)
            elif typ == Constants.ObjectTypes.COLLECTION:
                if not po in self.__deletedCollections:
                    self.__deletedCollections.append(po)
            else:
                raise "Unknown Object Type %s" % typ

            for eName in po._extents:
                if not self.__extentsDel.has_key(eName):
                    self.__extentsDel[eName] = []
                self.__extentsDel[eName].append(po)
        finally:
            self.__lk.release()


    def _4ods_registerLargeFileDelete(self,fileName):
        if not self.__open:
            raise TransactionNotInProgress()
        self.__largeFilesToDelete.append(fileName)

    def _4ods_registerLargeFileWrite(self,fileName,data):
        if not self.__open:
            raise TransactionNotInProgress()
        self.__largeFilesToWrite.append((fileName,data))
        





    #PAss through functions
    #We need to gaurentte that all of these happen in a single thread

    def _4ods_getRepositoryObject(self,rid):
        self.__lk.acquire()
        try:
            return self.__storageManager.getRepositoryObject(rid)
        finally:
            self.__lk.release()

    def _4ods_getObject(self,oid):
        self.__lk.acquire()
        try:
            return self.__storageManager.getObject(oid)
        finally:
            self.__lk.release()

    def _4ods_getCollection(self,cid):
        self.__lk.acquire()
        try:
            return self.__storageManager.getCollection(cid)
        finally:
            self.__lk.release()

    def _4ods_loadCollectionData(self,cid,index):
        self.__lk.acquire()
        try:
            return self.__storageManager.loadCollectionData(cid,index)
        finally:
            self.__lk.release()

    def _4ods_lookup(self,name):
        self.__lk.acquire()
        try:
            obj =  self.__storageManager.lookup(name)
            return obj
        finally:
            self.__lk.release()


    def _4ods_bind(self,object,name):
        self.__lk.acquire()
        try:
            return self.__storageManager.bind(object,name)
        finally:
            self.__lk.release()

    def _4ods_loadExtent(self,id):
        self.__lk.acquire()
        try:
            return self.__storageManager.loadExtent(id)
        finally:
            self.__lk.release()

    def _4ods_unbind(self,name):
        self.__lk.acquire()
        try:
            return self.__storageManager.unbind(name)
        finally:
            self.__lk.release()
    def _4ods_getAllBindings(self,object):
        self.__lk.acquire()
        try:
            return self.__storageManager.getAllBindings(object)
        finally:
            self.__lk.release()

    def _4ods_isBound(self,object):
        self.__lk.acquire()
        try:
            return self.__storageManager.isBound(object)
        finally:
            self.__lk.release()

    def _4ods_getNewObjectByExtent(self,name):
        self.__lk.acquire()
        try:
            return self.__extentsAdd.get(name,[])
        finally:
            self.__lk.release()
        
    def _4ods_getDeleteObjectByExtent(self,name):
        self.__lk.acquire()
        try:
            return self.__extentsDel.get(name,[])
        finally:
            self.__lk.release()
        
    def _4ods_runQuery(self,cmd):
        self.__lk.acquire()
        try:
            return self.__storageManager.runQuery(cmd)
        finally:
            self.__lk.release()
        
